;;; -*- Mode: Lisp; Package: OUTLINE-TREE2 -*-
;;; cr-regexp-inheritance.l --- outline-tree

;; Copyright (C) 2001-2006 OHKUBO Hiroshi <ohkubo@s53.xrea.com>

;; Author: OHKUBO Hiroshi <ohkubo@s53.xrea.com>
;; Time-stamp: <2006/08/25 12:07:17 +0900>

;;; Code:

(provide "outline-tree/cr-regexp-inheritance")

(require "outline-tree/outline-tree")
(require "outline-tree/cr-regexp-internal")

(in-package "outline-tree2")

;; s؂
(defun outline-tree-create-outline-line (root-node)
  (outline-tree-create-outline-regexp-internal
   :root-node root-node
   :layer-regexp-list '("^\\(.+$\\|\n\\)")
   :layer-regexp-list-case-fold nil
   :layer-regexp-list-order :non-miss
   :line-oriented t
   :title-converter #'(lambda (str) (if (string= "\n" str) "" str))))
(outline-tree-add-create-outline-function
 'outline-tree-create-outline-line "s؂(Ss)" '(:generic :line-oriented :sequential)
 "esSĕ\܂B
es̓ւɓKĂ܂B")

;; Cfgs؂
(defun outline-tree-create-outline-no-space-line (root-node)
  (outline-tree-create-outline-regexp-internal
   :root-node root-node
   :layer-regexp-list '("^[^ @\t\n].*$")
   :layer-regexp-list-case-fold nil
   :layer-regexp-list-order :non-miss
   :line-oriented t))
(outline-tree-add-create-outline-function
 'outline-tree-create-outline-no-space-line "s؂(Cfgs)" '(:generic :line-oriented :sequential)
 "sɃCfgȂsŋ؂܂B")

;; ob`t@C 
(defun outline-tree-create-outline-bat (root-node)
  (outline-tree-create-outline-regexp-internal
   :root-node root-node
   :layer-regexp-list '("^\\(:\\S+\\)$")
   :layer-regexp-list-case-fold t
   :layer-regexp-list-order :non-miss
   :line-oriented t))
(outline-tree-add-create-outline-function
 'outline-tree-create-outline-bat "BAT (x)" '(:misc :line-oriented :sequential)
 "ob`t@C (*.bat) s̃x (:***) ɕ\܂B")

;; LISP Rg 
(defun outline-tree-create-outline-lisp-comment (root-node)
  (outline-tree-create-outline-regexp-internal
   :root-node root-node
   :layer-regexp-list
   '("^;;;?"
     "^ *(\\(?:[a-zA-Z]+::?\\)?\\(?:defun\\|defmacro\\|defvar\\|defvar-local\\|defconstant\\|define-dll-entry\\)")
   :layer-regexp-list-case-fold nil
   :layer-regexp-list-order :top-miss))
(outline-tree-add-create-outline-function
 'outline-tree-create-outline-lisp-comment "LISP (Rg)" '(:language :line-oriented :sequential)
 "LISP t@Cs \";;\" Ŏn܂RgƁAe` S 
AEgC\܂B")

;; info tH[}bg
(defun outline-tree-create-outline-info (root-node)
  (outline-tree-create-outline-regexp-internal
   :root-node root-node
   :layer-regexp-list
   '("^\\([^ \t\n].+\\)\n\\*\\{2,\\}$"
     "^\\([^ \t\n].+\\)\n=\\{2,\\}$"
     "^\\([^ \t\n].+\\)\n-\\{2,\\}$"
     "^\\([^ \t\n].+\\)\n\\.\\{2,\\}$")
   :layer-regexp-list-case-fold nil
   :layer-regexp-list-order :top-miss
   :sub-type :heading-node))
(outline-tree-add-create-outline-function
 'outline-tree-create-outline-info "info" '(:format :line-oriented :sequential)
 "info tH[}bgAEgC\܂B")

;; ChangeLog tH[}bg
(defun outline-tree-create-outline-changelog (root-node)
  (outline-tree-create-outline-regexp-internal
   :root-node root-node
   :layer-regexp-list
   '("^\\([^ \t\n].+\\)"
     "^\t\\* \\(.+\\)")
   :layer-regexp-list-case-fold nil
   :layer-regexp-list-order :non-miss
   :sub-type :heading-node
   :line-oriented t))
(outline-tree-add-create-outline-function
 'outline-tree-create-outline-changelog "ChangeLog" '(:format :line-oriented :sequential)
 "ChangeLog tH[}bgAEgC\܂B")

;; roff tH[}bg
(defun outline-tree-create-outline-roff (root-node)
  (outline-tree-create-outline-regexp-internal
   :root-node root-node
   :layer-regexp-list
   '("^\\.TH \\(.+\\)$"
     "^\\.SH \"?\\(.+?\\)\"? *$"
     "^\\.[^ \n]+ \"?\\(.+?\\)\"? *$")
   :layer-regexp-list-case-fold nil
   :layer-regexp-list-order :any-miss
   :line-oriented t
   :sub-type :heading-node))
(outline-tree-add-create-outline-function
 'outline-tree-create-outline-roff "roff" '(:format :line-oriented :sequential)
 "roff tH[}bgAEgC\܂B")

;; Java 
(defun outline-tree-create-outline-java (root-node)
  (outline-tree-create-outline-regexp-internal
   :root-node root-node
   :layer-regexp-list
   '("^[ \t]*\\(\\(?:package\\|import\\|[^{\n]*?class[ \t]+[a-zA-Z]+\\).*\\)[ \t]*[;{]"
     "^[ \t]*\\([^{\n()]*?\\(?:public\\|private\\)[^{\n()]*?[a-zA-Z]+[ \t]*(.*)[^(){}]*\\)[ \t]*{")
   :layer-regexp-list-case-fold nil
   :layer-regexp-list-order :non-miss))
(outline-tree-add-create-outline-function
 'outline-tree-create-outline-java "Java" '(:language :line-oriented :sequential)
 "Java ɊȈՓIȕ@ŃAEgC\܂B
pɑςɂ͎Ă܂B")

;;; cr-regexp-inheritance.l ends here
